using System.Collections;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using Upsilon.BaseTypes.ScriptTypeInterfaces;
using Upsilon.BaseTypes.UserData;
using Upsilon.Binder;
using Upsilon.Binder.VariableSymbols;
using Upsilon.Evaluator;
using Upsilon.Exceptions;
using Upsilon.Text;

namespace Upsilon.BaseTypes.ScriptFunction
{
    internal class ScriptRuntimeFunction : ScriptFunction
    {
        public List<ScriptRuntimeFunctionOption> Options { get; }

        public ScriptRuntimeFunction(string name, List<ScriptRuntimeFunctionOption> options, bool isCoroutine) : base(name, isCoroutine)
        {
            Options = options;
        }

        public void SetName(string name)
        {
            Name = name;
        }

        public override ScriptType Run(Diagnostics diagnostics, ScriptType[] variables, EvaluationScope scope, TextSpan span, EvaluationState state)
        {
            var option = GetValidOption(variables);
            if (option == null)
            {
                throw new EvaluationException(state.Script.FileName,
                    $"No valid function found", span);
            }
            state.Stacktrace.Push(this);
            var result = option.Run(diagnostics, variables, state.Script, scope, span);
            state.Stacktrace.Pop();
            return result;
        }

        public override IEnumerator RunCoroutine(Diagnostics diagnostics, ScriptType[] variables, EvaluationScope scope,
            TextSpan span, EvaluationState state)
        {
            var option = GetValidOption(variables);
            if (option == null)
            {
                throw new EvaluationException(state.Script.FileName,
                    $"No valid function found", span);
            }
            state.Stacktrace.Push(this);
            var coroutine = option.RunCoroutine(diagnostics, variables, state.Script, scope, span);
            while (coroutine.MoveNext())
            {
                yield return coroutine.Current;
            }
            state.Stacktrace.Pop();
        }

        public ScriptRuntimeFunctionOption GetValidOption(object[] variables)
        {
            if (variables == null)
            {
                return Options.FirstOrDefault(x => x.Parameters.Length == 0);
            }
            foreach (var option in Options)
            {
                if (option.Parameters.Length != variables.Length)
                    continue;
                var isCompatible = true;
                for (var index = 0; index < variables.Length; index++)
                {
                    var parameter       = option.Parameters[index];
                    var v = variables[index];
                    if (v == null)
                    {
                        continue;
                    }
                    System.Type parameterType;
                    if (v is GenericUserData ud)
                        parameterType = ud.GetCSharpType();
                    else
                        parameterType = v.GetType();

                    if (parameter.VariableSymbol is UserDataVariableSymbol parameterSymbol)
                    {
                        if (parameterSymbol.BoundTypeDefinition != null)
                        {
                            var validSymbol =
                                parameterSymbol.BoundTypeDefinition.ValidInternalTypes.Any(validType =>
                                {
                                    var b = validType.IsAssignableFrom(parameterType);
                                    if (b) return true;
                                    if (parameterType.IsGenericType &&
                                        validType.IsAssignableFrom(parameterType.GetGenericTypeDefinition()))
                                        return true;
                                    return false;
                                });
                            if (!validSymbol)
                            {
                                isCompatible = false;
                                break;
                            }
                        }
                    }
                    else if (parameter.VariableSymbol is TableVariableSymbol)
                    {
                        if (parameterType.GetScriptType() != BaseTypes.Type.Table)
                        {
                            isCompatible = false;
                            break;
                        }
                    }
                }
                if (!isCompatible)
                    continue;
                return option;
            }
            return null;
        }

        public ScriptRuntimeFunctionOption GetValidOption(ScriptType[] variables)
        {
            foreach (var option in Options)
            {
                if (option.Parameters.Length != variables.Length)
                    continue;
                bool isCompatible = true;
                for (var index = 0; index < variables.Length; index++)
                {
                    var callingVariable = variables[index];
                    var optionVariable = option.Parameters[index];
                    if (callingVariable.Type == BaseTypes.Type.Unknown || callingVariable.Type == BaseTypes.Type.Nil ||
                        optionVariable.ValueType == BaseTypes.Type.Unknown)
                        continue;
                    if (callingVariable.Type != optionVariable.ValueType)
                    {
                        isCompatible = false;
                        break;
                    }
                }
                if (!isCompatible)
                    continue;
                return option;
            }
            return null;
        }

        public void AddOption(ScriptRuntimeFunctionOption newOption)
        {
            var variables = newOption.Parameters;
            for (var i = 0; i < Options.Count; i++)
            {
                var option = Options[i];
                if (option.Parameters.Length != variables.Length)
                    continue;
                bool isCompatible = true;
                for (var index = 0; index < variables.Length; index++)
                {
                    var callingVariable = variables[index];
                    var optionVariable  = option.Parameters[index];
                    if (callingVariable.ValueType == BaseTypes.Type.Unknown || callingVariable.ValueType == BaseTypes.Type.Nil ||
                        optionVariable.ValueType  == BaseTypes.Type.Unknown)
                        continue;
                    if (callingVariable.ValueType != optionVariable.ValueType)
                    {
                        isCompatible = false;
                        break;
                    }
                }
                if (!isCompatible)
                    continue;
                Options[i] = newOption;
                break;
            }
            Options.Add(newOption);
        }


        public class ScriptRuntimeFunctionOption : IScopeOwner
        {
            public ScriptRuntimeFunctionOption(ImmutableArray<BoundVariableSymbol> parameters,
            BoundBlockStatement block, EvaluationScope scope)
            {
                Parameters = parameters;
                Block = block;
                EvaluationScope = scope;
            }

            public EvaluationScope EvaluationScope { get; private set; }

            public BoundBlockStatement                 Block           { get; }
            public ImmutableArray<BoundVariableSymbol> Parameters      { get; }

            public ScriptType Run(Diagnostics diagnostics, ScriptType[] variables, Script script, EvaluationScope outerScope, TextSpan span)
            {
                var innerScope = new EvaluationScope(EvaluationScope);
                var state = new EvaluationState()
                {
                    Script = script,
                };
                if (Parameters != null)
                {
                    for (var i = 0; i < Parameters.Length; i++)
                    {
                        var parameterVariable = Parameters[i];
                        var parameterValue    = variables[i];
                        innerScope.CreateLocal(parameterVariable.VariableSymbol, parameterValue);
                    }
                }

                Block.Evaluate(innerScope, diagnostics, ref state);
                return state.ReturnValue;
            }

            public IEnumerator RunCoroutine(Diagnostics diagnostics, ScriptType[] variables, Script script, EvaluationScope outerScope, TextSpan span)
            {
                var innerScope = new EvaluationScope(EvaluationScope);
                var state = new EvaluationState()
                {
                    Script = script,
                };
                if (Parameters != null)
                {
                    for (var i = 0; i < Parameters.Length; i++)
                    {
                        var parameterVariable = Parameters[i];
                        var parameterValue    = variables[i];
                        innerScope.CreateLocal(parameterVariable.VariableSymbol, parameterValue);
                    }
                }

                return Block.EvaluateCoroutine(innerScope, diagnostics, state);
            }


        }
    }
}