using System;
using System.Collections.Generic;
using Upsilon.BaseTypes;
using Upsilon.BaseTypes.Number;
using Upsilon.BaseTypes.UserData;
using Upsilon.Evaluator;
using Upsilon.Exceptions;
using Upsilon.Text;
using Type = Upsilon.BaseTypes.Type;

namespace Upsilon.Binder
{
    public class BoundBinaryExpression : BoundExpression
    {
        public BoundBinaryExpression(BoundBinaryOperator op, BoundExpression leftExpression,
            BoundExpression rightExpression, Type valueType, TextSpan span) : base(span)
        {
            Operator = op;
            LeftExpression = leftExpression;
            RightExpression = rightExpression;
            ValueType = valueType;
        }

        public override BoundKind Kind => BoundKind.BoundBinaryExpression;
        public override IEnumerable<BoundNode> GetChildren()
        {
            yield return LeftExpression;
            yield return RightExpression;
        }

        public override TypeContainer ValueType { get; }

        internal override ScriptType Evaluate(EvaluationScope scope, Diagnostics diagnostics, ref EvaluationState state)
        {
            if (Operator.Kind == BoundBinaryOperator.OperatorKind.Or)
            {
                var l = LeftExpression.Evaluate(scope, diagnostics, ref state);
                if (l.Type == Type.Boolean && ((ScriptBoolean)l).Value)
                    return new ScriptBoolean(true);
                var r = RightExpression.Evaluate(scope, diagnostics, ref state);
                if (r.Type == Type.Boolean && ((ScriptBoolean)r).Value)
                    return new ScriptBoolean(true);
                return new ScriptBoolean(false);
            }
            if (Operator.Kind == BoundBinaryOperator.OperatorKind.And)
            {
                var l = LeftExpression.Evaluate(scope, diagnostics, ref state);
                if (l.Type != Type.Boolean || !((ScriptBoolean)l).Value)
                    return new ScriptBoolean(false);
                var r = RightExpression.Evaluate(scope, diagnostics, ref state);
                if (r.Type != Type.Boolean || !((ScriptBoolean)r).Value)
                    return new ScriptBoolean(false);
                return new ScriptBoolean(true);
            }

            var left = LeftExpression.Evaluate(scope, diagnostics, ref state);
            var right = RightExpression.Evaluate(scope, diagnostics, ref state);
            switch (Operator.Kind)
            {
                case BoundBinaryOperator.OperatorKind.Addition:
                    if (left.Type == Type.Number)
                    {
                        if (right.Type == Type.Number)
                        {
                            return ((ScriptNumber)left) + ((ScriptNumber)right);
                        }
                        if (right.Type == Type.String)
                        {
                            return new ScriptString(left + right.ToString());
                        }
                    }
                    else if (left.Type == Type.String)
                    {
                        return ((ScriptString) left) + right;
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Addition, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Subtraction:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) - ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Subtraction, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Multiplication:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) * ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Multiplication, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Division:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) / ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Division, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Exponent:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ScriptNumber.Exponent((ScriptNumber) left, (ScriptNumber) right);
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Remainder:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) % ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Modulo, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Equality:
                    return new ScriptBoolean(left.Equals(right));
                case BoundBinaryOperator.OperatorKind.Inequality:
                    return new ScriptBoolean(!left.Equals(right));
                case BoundBinaryOperator.OperatorKind.Less:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber) left) < ((ScriptNumber) right);
                    }

                    throw new EvaluationException(state.Script.FileName,
                        $"Can't find operator '{Operator.Kind}' for types '{left.Type}' and '{right.Type}'", Span);
                case BoundBinaryOperator.OperatorKind.LessEquals:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber) left) <= ((ScriptNumber) right);
                    }

                    throw new EvaluationException(state.Script.FileName,
                        $"Can't find operator '{Operator.Kind}' for types '{left.Type}' and '{right.Type}'", Span);
                case BoundBinaryOperator.OperatorKind.Greater:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber) left) > ((ScriptNumber) right);
                    }

                    throw new EvaluationException(state.Script.FileName,
                        $"Can't find operator '{Operator.Kind}' for types '{left.Type}' and '{right.Type}'", Span);
                case BoundBinaryOperator.OperatorKind.GreaterEquals:
                    if (left.Type == Type.Number && right.Type == Type.Number)
                    {
                        return ((ScriptNumber) left) >= ((ScriptNumber) right);
                    }

                    throw new EvaluationException(state.Script.FileName,
                        $"Can't find operator '{Operator.Kind}' for types '{left.Type}' and '{right.Type}'", Span);
                default:
                    throw new Exception("Invalid Binary Operator: " + Operator.Kind);
            }
        }

        public BoundBinaryOperator Operator { get; }
        public BoundExpression LeftExpression { get; }
        public BoundExpression RightExpression { get; }
    }
}