using System.Collections.Generic;
using System.Collections.Immutable;
using Upsilon.BaseTypes;
using Upsilon.BaseTypes.ScriptTable;
using Upsilon.Binder.VariableSymbols;
using Upsilon.Evaluator;
using Upsilon.Text;
using Type = Upsilon.BaseTypes.Type;

namespace Upsilon.Binder
{
    public class BoundTableExpression : BoundExpression, IBoundNodeScopeOwner
    {
        public BoundTableExpression(Dictionary<string, VariableSymbol> expressions,
            ImmutableArray<BoundStatement> statements, TextSpan span, BoundScope scope) : base(span)
        {
            Expressions = expressions;
            Statements = statements;
            Scope = scope;
        }

        public override BoundKind Kind => BoundKind.BoundTableExpression;

        public override IEnumerable<BoundNode> GetChildren()
        {
            return Statements;
        }

        public override TypeContainer ValueType
        {
            get
            {
                TypeContainer valueType = null;
                foreach (var statement in Statements)
                {
                    if (!(statement is BoundExpressionStatement exp))
                    {
                        valueType = BaseTypes.Type.Unknown;
                        break;
                    }
                    if (valueType == null)
                    {
                        valueType = exp.Expression.ValueType;
                        continue;
                    }
                    if (valueType == exp.Expression.ValueType) continue;
                    valueType = BaseTypes.Type.Unknown;
                    break;
                }

                TypeContainer valueRealType = BaseTypes.Type.Unknown;
                if (valueType != null)
                    valueRealType = valueType;

                var arr = new TypeContainer[] {BaseTypes.Type.String, valueRealType};
                return new CompositeTypeContainer(arr.ToImmutableArray());
            }
        }

        internal override ScriptType Evaluate(EvaluationScope outerScope, Diagnostics diagnostics, ref EvaluationState outerState)
        {
            var tableScope     = EvaluationScope.CreateWithGetOnlyParent(outerScope);
            var innerState = new EvaluationState()
            {
                Script = outerState.Script
            };


            var currentPos     = 1;
            var isNumerated    = true;
            foreach (var boundStatement in Statements)
            {
                switch (boundStatement.Kind)
                {
                    case BoundKind.BoundAssignmentStatement:
                    case BoundKind.BoundFunctionExpression:
                    case BoundKind.BoundFunctionAssignmentStatement:
                        boundStatement.Evaluate(tableScope, diagnostics, ref innerState);
                        isNumerated = false;
                        break;
                    default:
                    {
                        boundStatement.Evaluate(tableScope, diagnostics, ref innerState);
                        if (innerState.LastValue != null)
                        {
                            tableScope.CreateLocal(new VariableSymbol(currentPos.ToString(), innerState.LastValue.Type, false),
                                innerState.LastValue);
                        }
                        innerState.LastValue = null;
                        break;
                    }
                }

                currentPos++;
            }

            if (isNumerated)
            {
                return new NumeratedScriptTable(tableScope);
            }
            return new GenericKeyedScriptTable(tableScope);

        }


        public Dictionary<string, VariableSymbol> Expressions { get; }
        public ImmutableArray<BoundStatement> Statements { get; }
        public BoundScope Scope { get; set; }
    }
}