using System.Collections.Generic;
using Upsilon.BaseTypes;
using Upsilon.BaseTypes.Number;
using Upsilon.BaseTypes.ScriptTypeInterfaces;
using Upsilon.BaseTypes.UserData;
using Upsilon.Evaluator;
using Upsilon.Exceptions;
using Upsilon.Text;

namespace Upsilon.Binder
{
    public class BoundUnaryExpression : BoundExpression
    {
        public override BoundKind Kind => BoundKind.BoundUnaryExpression;

        public override IEnumerable<BoundNode> GetChildren()
        {
            yield return InExpression;
        }

        public override TypeContainer ValueType { get; }
        internal override ScriptType Evaluate(EvaluationScope scope, Diagnostics diagnostics, ref EvaluationState state)
        {
            var operand = InExpression.Evaluate(scope, diagnostics, ref state);
            switch (Operator.Kind)
            {
                case BoundUnaryOperator.OperatorKind.Identity:
                    return operand;
                case BoundUnaryOperator.OperatorKind.Negation:
                    if (operand.Type == Type.Number)
                        return -((ScriptNumber)operand);
                    else if (operand.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) operand;
                        var (type, failed) = ud.UnaryOperator(operand, OperatorType.UnaryNegation);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundUnaryOperator.OperatorKind.LogicalNegation:
                    if (operand.Type == Type.Boolean)
                        return !(ScriptBoolean) operand;
                    else if (operand.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) operand;
                        var (type, failed) = ud.UnaryOperator(operand, OperatorType.LogicalNot);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundUnaryOperator.OperatorKind.Length:
                    if (operand is ILengthType length)
                    {
                        return length.Length();
                    }
                    goto default;
                default:
                    throw new EvaluationException(state.Script.FileName, "Invalid Unary Operator: " + Operator.Kind, Span);
            }

        }

        public BoundUnaryExpression(BoundUnaryOperator op, BoundExpression inExpression, Type type, TextSpan span) : base(span)
        {
            Operator        = op;
            InExpression = inExpression;
            ValueType            = type;
        }

        public BoundExpression InExpression { get; }
        public BoundUnaryOperator Operator { get; }
    }
}