using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using System.Reflection;
using Upsilon.BaseTypes;
using Upsilon.BoundTypes;
using Type = Upsilon.BaseTypes.Type;

namespace Upsilon.Binder.VariableSymbols
{
    public class InternalFunctionVariableSymbol : FunctionVariableSymbol
    {
        public InternalFunctionVariableSymbol(string name, bool local, TypeContainer resultType,
            InternalFunctionParameter[] functionParameters, MethodInfo overrideResultType, bool isCoroutine)
            : base(name, local, isCoroutine)
        {
            FunctionOption.Add(new InternalFunctionVariableOption(resultType, functionParameters, overrideResultType));
        }

        public InternalFunctionVariableSymbol(string name, bool local, List<FunctionVariableSymbolOption> options,
            bool isCoroutine)
            : base(name, local, isCoroutine)
        {
            FunctionOption = options;
        }

        public override bool ValidateParameters(ImmutableArray<BoundExpression> callingParameters)
        {
            foreach (var functionVariableSymbolOption in FunctionOption)
            {
                if (!(functionVariableSymbolOption is InternalFunctionVariableOption option))
                    continue;
                if (option.FunctionParameters.Length != callingParameters.Length)
                    continue;
                var isValid = true;
                for (var i = 0; i < callingParameters.Length; i++)
                {
                    var functionParameter = option.FunctionParameters[i];
                    var callingParameter  = callingParameters[i];
                    if (callingParameter.ValueType == Type.Unknown || callingParameter.ValueType == Type.Nil)
                    {
                        continue;
                    }
                    if (functionParameter.ValidTypes == Type.Unknown)
                        continue;

                    if (!functionParameter.ValidTypes.Type.HasFlag(callingParameter.ValueType))
                    {
                        isValid = false;
                        break;
                    }

                    if (functionParameter.ValidTypes.Type.HasFlag(Type.UserData))
                    {
                        var variable = Binder.ResolveVariable(callingParameter, null);
                        if (variable != null && variable.TypeContainer == Type.UserData)
                        {
                            var parent =
                                (UserDataBoundTypeDefinition) ((UserDataVariableSymbol) variable).BoundTypeDefinition;
                            if (functionParameter.ExpectedUserData != null && !functionParameter.ExpectedUserData.Contains(parent.Name))
                            {
                                isValid = false;
                                break;
                            }
                        }
                    }
                }
                if (!isValid)
                    continue;
                return true;
            }

            return false;
        }

        public TypeContainer GetResultType(BoundExpression[] parameters)
        {
            foreach (var functionVariableSymbolOption in FunctionOption)
            {
                if (!(functionVariableSymbolOption is InternalFunctionVariableOption option))
                    continue;
                if (option.FunctionParameters.Length != parameters.Length)
                    continue;
                var isValid = true;
                for (var i = 0; i < parameters.Length; i++)
                {
                    var functionParameter = option.FunctionParameters[i];
                    var callingParameter  = parameters[i];
                    if (!functionParameter.ValidTypes.Type.HasFlag(callingParameter.ValueType))
                    {
                        isValid = false;
                        break;
                    }
                }
                if (!isValid)
                    continue;
                if (option.OverrideResultType == null)
                    return option.ResultType;
                return (TypeContainer) option.OverrideResultType.Invoke(null, new object[] {parameters});
            }
            return Type.Unknown;
        }

        public class InternalFunctionParameter
        {
            public InternalFunctionParameter(string name, TypeContainer type, bool isOptional)
            {
                ValidTypes = type;
                IsOptional = isOptional;
                Name       = name;
            }

            public InternalFunctionParameter(string name, TypeContainer type, string[] expectedUserData, bool isOptional)
            {
                ValidTypes       = type;
                ExpectedUserData = expectedUserData;
                IsOptional       = isOptional;
                Name             = name;
            }

            public string   Name             { get; }
            public TypeContainer     ValidTypes       { get; }
            public string[] ExpectedUserData { get; }
            public bool     IsOptional       { get; }
        }
    }

    public class InternalFunctionVariableOption : FunctionVariableSymbolOption
    {
        public InternalFunctionVariableSymbol.InternalFunctionParameter[] FunctionParameters { get; }
        public MethodInfo OverrideResultType { get; }

        public InternalFunctionVariableOption(TypeContainer resultType,
            InternalFunctionVariableSymbol.InternalFunctionParameter[] functionParameters,
            MethodInfo overrideResultType) : base(resultType)
        {
            FunctionParameters = functionParameters;
            OverrideResultType = overrideResultType;
        }

        public override TypeContainer[] GetParameterTypes()
        {
            return FunctionParameters.Select(x => x.ValidTypes).ToArray();
        }
    }
}